@@ -26,6 +26,7 @@
 #include <string>
 #include <vector>
 
+#include "fileread.h"
 #include "workqueue.h"
 
 namespace gold
@@ -37,6 +38,8 @@ class Input_objects;
 class Input_group;
 class Layout;
 class Symbol_table;
+class Object;
+class Read_symbols_data;
 
 // This class represents an archive--generally a libNAME.a file.
 // Archives have a symbol table and a list of objects.
@@ -47,8 +50,9 @@ class Archive
   Archive(const std::string& name, Input_file* input_file,
           bool is_thin_archive, Dirsearch* dirpath, Task* task)
     : name_(name), input_file_(input_file), armap_(), armap_names_(),
-      extended_names_(), armap_checked_(), seen_offsets_(),
-      is_thin_archive_(is_thin_archive), dirpath_(dirpath), task_(task)
+      extended_names_(), armap_checked_(), seen_offsets_(), members_(),
+      is_thin_archive_(is_thin_archive), nested_archives_(),
+      dirpath_(dirpath), task_(task), num_members_(0)
   { }
 
   // The length of the magic string at the start of an archive.
@@ -61,20 +65,31 @@ class Archive
   // The string expected at the end of an archive member header.
   static const char arfmag[2];
 
-  // The name of the object.
+  // The name of the object.  This is the name used on the command
+  // line; e.g., if "-lgcc" is on the command line, this will be
+  // "gcc".
   const std::string&
   name() const
   { return this->name_; }
 
+  // The file name.
+  const std::string&
+  filename() const
+  { return this->input_file_->filename(); }
+
   // Set up the archive: read the symbol map.
   void
-  setup();
+  setup(Input_objects*);
 
   // Get a reference to the underlying file.
   File_read&
   file()
   { return this->input_file_->file(); }
 
+  const File_read&
+  file() const
+  { return this->input_file_->file(); }
+
   // Lock the underlying file.
   void
   lock(const Task* t)
@@ -105,6 +120,11 @@ class Archive
   clear_uncached_views()
   { this->input_file_->file().clear_uncached_views(); }
 
+  // Whether this is a thin archive.
+  bool
+  is_thin_archive() const
+  { return this->is_thin_archive_; }
+
   // Unlock any nested archives.
   void
   unlock_nested_archives();
@@ -112,7 +132,15 @@ class Archive
   // Select members from the archive as needed and add them to the
   // link.
   void
-  add_symbols(Symbol_table*, Layout*, Input_objects*);
+  add_symbols(Symbol_table*, Layout*, Input_objects*, Mapfile*);
+
+  // Dump statistical information to stderr.
+  static void
+  print_stats();
+
+  // Return the number of members in the archive.
+  size_t
+  count_members();
 
  private:
   Archive(const Archive&);
@@ -120,6 +148,13 @@ class Archive
 
   struct Archive_header;
 
+  // Total number of archives seen.
+  static unsigned int total_archives;
+  // Total number of archive members seen.
+  static unsigned int total_members;
+  // Number of archive members loaded.
+  static unsigned int total_members_loaded;
+
   // Get a view into the underlying file.
   const unsigned char*
   get_view(off_t start, section_size_type size, bool aligned, bool cache)
@@ -139,15 +174,51 @@ class Archive
   // member, and set *PNAME to the name.
   off_t
   interpret_header(const Archive_header* hdr, off_t off, std::string* pname,
-                   off_t* nested_off);
+                   off_t* nested_off) const;
+
+  // Get the file and offset for an archive member, which may be an
+  // external member of a thin archive.  Set *INPUT_FILE to the
+  // file containing the actual member, *MEMOFF to the offset
+  // within that file (0 if not a nested archive), and *MEMBER_NAME
+  // to the name of the archive member.  Return TRUE on success.
+  bool
+  get_file_and_offset(off_t off, Input_objects* input_objects,
+                      Input_file** input_file, off_t* memoff,
+                      std::string* member_name);
+
+  // Return an ELF object for the member at offset OFF.  Set *MEMBER_NAME to
+  // the name of the member.
+  Object*
+  get_elf_object_for_member(off_t off, Input_objects* input_objects);
+
+  // Read the symbols from all the archive members in the link.
+  void
+  read_all_symbols(Input_objects* input_objects);
+
+  // Read the symbols from an archive member in the link.  OFF is the file
+  // offset of the member header.
+  void
+  read_symbols(Input_objects* input_objects, off_t off);
 
   // Include all the archive members in the link.
   void
-  include_all_members(Symbol_table*, Layout*, Input_objects*);
+  include_all_members(Symbol_table*, Layout*, Input_objects*, Mapfile*);
 
   // Include an archive member in the link.
   void
-  include_member(Symbol_table*, Layout*, Input_objects*, off_t off);
+  include_member(Symbol_table*, Layout*, Input_objects*, off_t off,
+		 Mapfile*, Symbol*, const char* why);
+
+  // Iterate over archive members.
+  class const_iterator;
+
+  const_iterator
+  begin();
+
+  const_iterator
+  end();
+
+  friend class const_iterator;
 
   // An entry in the archive map of symbols to object files.
   struct Armap_entry
@@ -158,6 +229,21 @@ class Archive
     off_t file_offset;
   };
 
+  // An entry in the archive map of offsets to members.
+  struct Archive_member
+  {
+    Archive_member()
+      : obj_(NULL), sd_(NULL)
+    { }
+    Archive_member(Object* obj, Read_symbols_data* sd)
+      : obj_(obj), sd_(sd)
+    { }
+    // The object file.
+    Object* obj_;
+    // The data to pass from read_symbols() to add_symbols().
+    Read_symbols_data* sd_;
+  };
+
   // A simple hash code for off_t values.
   class Seen_hash
   {
@@ -184,6 +270,8 @@ class Archive
   std::vector<bool> armap_checked_;
   // Track which elements have been included by offset.
   Unordered_set<off_t, Seen_hash> seen_offsets_;
+  // Table of objects whose symbols have been pre-read.
+  std::map<off_t, Archive_member> members_;
   // True if this is a thin archive.
   const bool is_thin_archive_;
   // Table of nested archives, indexed by filename.
@@ -192,6 +280,8 @@ class Archive
   Dirsearch* dirpath_;
   // The task reading this archive.
   Task *task_;
+  // Number of members in this archive;
+  unsigned int num_members_;
 };
 
 // This class is used to read an archive and pick out the desired
@@ -201,12 +291,12 @@ class Add_archive_symbols : public Task
 {
  public:
   Add_archive_symbols(Symbol_table* symtab, Layout* layout,
-		      Input_objects* input_objects,
+		      Input_objects* input_objects, Mapfile* mapfile,
 		      Archive* archive, Input_group* input_group,
 		      Task_token* this_blocker,
 		      Task_token* next_blocker)
     : symtab_(symtab), layout_(layout), input_objects_(input_objects),
-      archive_(archive), input_group_(input_group),
+      mapfile_(mapfile), archive_(archive), input_group_(input_group),
       this_blocker_(this_blocker), next_blocker_(next_blocker)
   { }
 
@@ -235,6 +325,7 @@ class Add_archive_symbols : public Task
   Symbol_table* symtab_;
   Layout* layout_;
   Input_objects* input_objects_;
+  Mapfile* mapfile_;
   Archive* archive_;
   Input_group* input_group_;
   Task_token* this_blocker_;
